@@ -209,12 +209,6 @@ static int set_up_temporary_mappings(void)
 	return 0;
 }
 
-#ifdef CONFIG_SUSPEND2
-void suspend2_beep(int suspending);
-#else
-#define suspend2_beep(num) do { } while(0)
-#endif
-
 int swsusp_arch_resume(void)
 {
 	int error;
@@ -222,7 +216,6 @@ int swsusp_arch_resume(void)
 	/* We have got enough memory and from now on we cannot recover */
 	if ((error = set_up_temporary_mappings()))
 		return error;
-	suspend2_beep(0);
 	restore_image();
 	return 0;
 }
