@@ -1,7 +1,7 @@
 /*
  * kernel/power/power_off.h
  *
- * Copyright (C) 2006 Nigel Cunningham <nigel@suspend2.net>
+ * Copyright (C) 2006-2007 Nigel Cunningham (nigel at suspend2 net)
  *
  * This file is released under the GPLv2.
  *
@@ -9,5 +9,5 @@
  */
 
 int suspend_pm_state_finish(void);
-void suspend_power_down(void);
-extern unsigned long suspend_powerdown_method;
+void suspend2_power_down(void);
+extern unsigned long suspend2_poweroff_method;
