@@ -1,7 +1,7 @@
 /*
  * kernel/power/suspend_block_io.c
  *
- * Copyright 2004-2006 Nigel Cunningham <nigel@suspend2.net>
+ * Copyright (C) 2004-2007 Nigel Cunningham (nigel at suspend2 net)
  *
  * Distributed under GPLv2.
  * 
@@ -22,52 +22,29 @@
 #include "block_io.h"
 #include "ui.h"
 
-/* Bits in struct io_info->flags */
-enum {
-	IO_AWAITING_READ,
-	IO_AWAITING_SUBMIT,
-	IO_AWAITING_CLEANUP,
-};
+static int pr_index;
+
+#if 0
+#define PR_DEBUG(a, b...) do { if (pr_index < 20) printk(a, ##b); } while(0)
+#else
+#define PR_DEBUG(a, b...) do { } while(0)
+#endif
 
 #define MAX_OUTSTANDING_IO 2048
+#define SUBMIT_BATCH_SIZE 128
 
-/*
- *
- *     IO in progress information storage and helpers
- *
- */
+static int max_outstanding_io = MAX_OUTSTANDING_IO;
+static int submit_batch_size = SUBMIT_BATCH_SIZE;
 
 struct io_info {
 	struct bio *sys_struct;
-	sector_t block[MAX_BUF_PER_PAGE];
-	struct page *buffer_page;
-	struct page *data_page;
-	unsigned long flags;
+	sector_t first_block;
+	struct page *bio_page, *dest_page;
+	int writing, readahead_index;
 	struct block_device *dev;
 	struct list_head list;
-	int readahead_index;
 };
 
-/*
- * submit_params
- */
-struct submit_params {
-	swp_entry_t swap_address;
-	struct page *page;
-	struct block_device *dev;
-	sector_t block[MAX_BUF_PER_PAGE];
-	int readahead_index;
-	struct submit_params *next;
-	int printme;
-};
-
-/* Locks separated to allow better SMP support.
- * An io_struct moves through the lists as follows.
- * free -> submit_batch -> busy -> ready_for_cleanup -> free
- */
-static LIST_HEAD(ioinfo_free);
-static DEFINE_SPINLOCK(ioinfo_free_lock);
-
 static LIST_HEAD(ioinfo_ready_for_cleanup);
 static DEFINE_SPINLOCK(ioinfo_ready_lock);
 
@@ -77,28 +54,24 @@ static DEFINE_SPINLOCK(ioinfo_submit_lock);
 static LIST_HEAD(ioinfo_busy);
 static DEFINE_SPINLOCK(ioinfo_busy_lock);
 
+static struct io_info *waiting_on;
+
 static atomic_t submit_batch;
-static int submit_batch_size = 64;
 static int submit_batched(void);
 
 /* [Max] number of I/O operations pending */
 static atomic_t outstanding_io;
-static int max_outstanding_io = 0;
-static atomic_t buffer_allocs, buffer_frees;
-
-/* [Max] number of pages used for above struct */
-static int infopages = 0;
-static int maxinfopages = 0;
 
 static int extra_page_forward = 0;
 
 static volatile unsigned long suspend_readahead_flags[
-	(MAX_OUTSTANDING_IO + BITS_PER_LONG - 1) / BITS_PER_LONG];
+	DIV_ROUND_UP(MAX_OUTSTANDING_IO, BITS_PER_LONG)];
 static spinlock_t suspend_readahead_flags_lock = SPIN_LOCK_UNLOCKED;
 static struct page *suspend_readahead_pages[MAX_OUTSTANDING_IO];
 static int readahead_index, readahead_submit_index;
 
 static int current_stream;
+/* 0 = Header, 1 = Pageset1, 2 = Pageset2 */
 struct extent_iterate_saved_state suspend_writer_posn_save[3];
 
 /* Pointer to current entry being loaded/saved. */
@@ -111,73 +84,11 @@ int suspend_writer_buffer_posn;
 
 int suspend_read_fd;
 
-static unsigned long nr_schedule_calls[8];
-
-static char *sch_caller[] = {
-	"get_io_info_struct #1    ",
-	"get_io_info_struct #2    ",
-	"get_io_info_struct #3    ",
-	"suspend_finish_all_io    ",
-	"wait_on_one_page         ",
-	"submit                   ",
-	"start_one                ",
-	"suspend_wait_on_readahead",
-};
-
 static struct suspend_bdev_info *suspend_devinfo;
 
 int suspend_header_bytes_used = 0;
 
-/*
- * suspend_reset_io_stats
- *
- * Description:	Reset all our sanity-checking statistics.
- */
-static void suspend_reset_io_stats(void)
-{
-	int i;
-	
-	max_outstanding_io = 0;
-	maxinfopages = 0;
-	
-	for (i = 0; i < 8; i++)
-		nr_schedule_calls[i] = 0;
-}
-
-/*
- * suspend_check_io_stats
- *
- * Description:	Check that our statistics look right and print
- * 		any debugging info wanted.
- */
-static void suspend_check_io_stats(void)
-{
-	int i;
-
-	BUG_ON(atomic_read(&outstanding_io));
-	BUG_ON(infopages);
-	BUG_ON(!list_empty(&ioinfo_submit_batch));
-	BUG_ON(!list_empty(&ioinfo_busy));
-	BUG_ON(!list_empty(&ioinfo_ready_for_cleanup));
-	BUG_ON(!list_empty(&ioinfo_free));
-	BUG_ON(atomic_read(&buffer_allocs) != atomic_read(&buffer_frees));
-
-	suspend_message(SUSPEND_WRITER, SUSPEND_LOW, 0,
-			"Maximum outstanding_io was %d.\n",
-			max_outstanding_io);
-	suspend_message(SUSPEND_WRITER, SUSPEND_LOW, 0,
-			"Max info pages was %d.\n",
-			maxinfopages);
-	if (atomic_read(&buffer_allocs) != atomic_read(&buffer_frees))
-		suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
-			"Buffer allocs (%d) != buffer frees (%d)",
-				atomic_read(&buffer_allocs),
-				atomic_read(&buffer_frees));
-	for(i = 0; i < 8; i++)
-		suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
-			"Nr schedule calls %s: %lu.\n", sch_caller[i],
-			nr_schedule_calls[i]);
-}
+DEFINE_MUTEX(suspend_bio_mutex);
 
 /*
  * __suspend_bio_cleanup_one
@@ -187,75 +98,50 @@ static void suspend_check_io_stats(void)
  */
 static void __suspend_bio_cleanup_one(struct io_info *io_info)
 {
-	struct page *buffer_page;
-	struct page *data_page;
-	char *buffer_address, *data_address;
-	int reading;
-
-	buffer_page = io_info->buffer_page;
-	data_page = io_info->data_page;
-
-	reading = test_bit(IO_AWAITING_READ, &io_info->flags);
 	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
-		"Cleanup IO: [%p]\n", 
-		io_info);
+		"Cleanup IO: [%p]\n", io_info);
 
-	if (reading && io_info->readahead_index == -1) {
+	if (!io_info->writing && io_info->readahead_index == -1) {
+		char *from, *to;
 		/*
 		 * Copy the page we read into the buffer our caller provided.
 		 */
-		data_address = (char *) kmap(data_page);
-		buffer_address = (char *) kmap(buffer_page);
-		memcpy(data_address, buffer_address, PAGE_SIZE);
-		kunmap(data_page);
-		kunmap(buffer_page);
-	
+		to = (char *) kmap(io_info->dest_page);
+		from = (char *) kmap(io_info->bio_page);
+		memcpy(to, from, PAGE_SIZE);
+		kunmap(io_info->dest_page);
+		kunmap(io_info->bio_page);
 	}
 
-	if (!reading || io_info->readahead_index == -1) {
+	if (io_info->writing || io_info->readahead_index == -1) {
 		/* Sanity check */
-		if (page_count(buffer_page) != 2)
+		if (page_count(io_info->bio_page) != 2)
 			printk(KERN_EMERG "Cleanup IO: Page count on page %p"
 					" is %d. Not good!\n",
-					buffer_page, page_count(buffer_page));
-		put_page(buffer_page);
-		__free_page(buffer_page);
-		atomic_inc(&buffer_frees);
+					io_info->bio_page,
+					page_count(io_info->bio_page));
+		put_page(io_info->bio_page);
+		__free_page(io_info->bio_page);
 	} else
-		put_page(buffer_page);
+		put_page(io_info->bio_page);
 	
 	bio_put(io_info->sys_struct);
 	io_info->sys_struct = NULL;
-	io_info->flags = 0;
 }
 
 /* __suspend_io_cleanup
  */
 
-static int suspend_bio_cleanup_one(void *data)
+static void suspend_bio_cleanup_one(void *data)
 {
 	struct io_info *io_info = (struct io_info *) data;
 	int readahead_index;
 	unsigned long flags;
 
-	/*
-	 * If this I/O was a readahead, remember its index.
-	 */
 	readahead_index = io_info->readahead_index;
-
-	/*
-	 * Add it to the free list.
-	 */
 	list_del_init(&io_info->list);
-	
-	/*
-	 * Do the cleanup.
-	 */
 	__suspend_bio_cleanup_one(io_info);
 
-	/*
-	 * Record the readahead as done.
-	 */
 	if (readahead_index > -1) {
 		int index = readahead_index/BITS_PER_LONG;
 		int bit = readahead_index - (index * BITS_PER_LONG);
@@ -264,15 +150,10 @@ static int suspend_bio_cleanup_one(void *data)
 		spin_unlock_irqrestore(&suspend_readahead_flags_lock, flags);
 	}
 
-	spin_lock_irqsave(&ioinfo_free_lock, flags);
-	list_add_tail(&io_info->list, &ioinfo_free);
-	spin_unlock_irqrestore(&ioinfo_free_lock, flags);
-	
-	/* Important: Must be last thing we do to avoid a race with
-	 * finish_all_io when using keventd to do the cleanup */
+	if (waiting_on == io_info)
+		waiting_on = NULL;
+	kfree(io_info);
 	atomic_dec(&outstanding_io);
-
-	return 0;
 }
 
 /* suspend_cleanup_some_completed_io
@@ -288,21 +169,15 @@ static void suspend_cleanup_some_completed_io(void)
 
 	spin_lock_irqsave(&ioinfo_ready_lock, flags);
 	while(!list_empty(&ioinfo_ready_for_cleanup)) {
-		int result;
 		first = list_entry(ioinfo_ready_for_cleanup.next,
 				struct io_info, list);
 
-		BUG_ON(!test_and_clear_bit(IO_AWAITING_CLEANUP, &first->flags));
-
 		list_del_init(&first->list);
 
 		spin_unlock_irqrestore(&ioinfo_ready_lock, flags);
-
-		result = suspend_bio_cleanup_one((void *) first);
-
+		suspend_bio_cleanup_one((void *) first);
 		spin_lock_irqsave(&ioinfo_ready_lock, flags);
-		if (result)
-			continue;
+
 		num_cleaned++;
 		if (num_cleaned == submit_batch_size)
 			break;
@@ -317,12 +192,10 @@ static void suspend_cleanup_some_completed_io(void)
  * Submit any I/O that's batched up (if we're not already doing
  * that, unplug queues, schedule and clean up whatever we can.
  */
-static void do_bio_wait(int caller)
+static void do_bio_wait(void)
 {
 	int num_submitted = 0;
 
-	nr_schedule_calls[caller]++;
-	
 	/* Don't want to wait on I/O we haven't submitted! */
 	num_submitted = submit_batched();
 
@@ -340,49 +213,9 @@ static void do_bio_wait(int caller)
  */
 static void suspend_finish_all_io(void)
 {
-	struct io_info *this, *next = NULL;
-	unsigned long flags;
-
 	/* Wait for all I/O to complete. */
 	while (atomic_read(&outstanding_io))
-		do_bio_wait(2);
-
-	spin_lock_irqsave(&ioinfo_free_lock, flags);
-	
-	/* 
-	 * Two stages, to avoid using freed pages.
-	 *
-	 * First free all io_info structs on a page except the first.
-	 */
-	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
-		if (((unsigned long) this) & ~PAGE_MASK)
-			list_del(&this->list);
-	}
-
-	/* 
-	 * Now we have only one reference to each page, and can safely
-	 * free pages, knowing we're not going to be trying to access the
-	 * same page after freeing it.
-	 */
-	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
-		list_del(&this->list);
-		free_page((unsigned long) this);
-		infopages--;
-		suspend_message(SUSPEND_MEMORY, SUSPEND_VERBOSE, 0,
-				"[FreedIOPage %lx]", this);
-	}
-	
-	spin_unlock_irqrestore(&ioinfo_free_lock, flags);
-}
-
-/*
- * wait_on_one_page
- *
- * Description:	Wait for a particular I/O to complete.
- */
-static void wait_on_one_page(struct io_info *io_info)
-{
-	do { do_bio_wait(3); } while (io_info->flags);
+		do_bio_wait();
 }
 
 /*
@@ -397,7 +230,7 @@ static void suspend_wait_on_readahead(int readahead_index)
 
 	/* read_ahead_index is the one we want to return */
 	while (!test_bit(bit, &suspend_readahead_flags[index]))
-		do_bio_wait(6);
+		do_bio_wait();
 }
 
 /*
@@ -418,7 +251,7 @@ static int suspend_readahead_ready(int readahead_index)
  * Set up for doing readahead on an image */
 static int suspend_prepare_readahead(int index)
 {
-	unsigned long new_page = get_zeroed_page(GFP_ATOMIC);
+	unsigned long new_page = get_zeroed_page(GFP_ATOMIC | __GFP_NOWARN);
 
 	if(!new_page)
 		return -ENOMEM;
@@ -455,8 +288,6 @@ static int suspend_end_bio(struct bio *bio, unsigned int num, int err)
 	list_del_init(&io_info->list);
 	spin_unlock_irqrestore(&ioinfo_busy_lock, flags);
 
-	set_bit(IO_AWAITING_CLEANUP, &io_info->flags);
-		
 	spin_lock_irqsave(&ioinfo_ready_lock, flags);
 	list_add_tail(&io_info->list, &ioinfo_ready_for_cleanup);
 	spin_unlock_irqrestore(&ioinfo_ready_lock, flags);
@@ -465,7 +296,7 @@ static int suspend_end_bio(struct bio *bio, unsigned int num, int err)
 
 /**
  *	submit - submit BIO request.
- *	@rw:	READ or WRITE.
+ *	@writing: READ or WRITE.
  *	@io_info: IO info structure.
  *
  * 	Based on Patrick's pmdisk code from long ago:
@@ -477,42 +308,40 @@ static int suspend_end_bio(struct bio *bio, unsigned int num, int err)
  *	Caller has already checked that our page is not fragmented.
  */
 
-static int submit(int rw, struct io_info *io_info)
+static int submit(struct io_info *io_info)
 {
-	int error = 0;
 	struct bio *bio = NULL;
 	unsigned long flags;
 
 	while (!bio) {
 		bio = bio_alloc(GFP_ATOMIC,1);
 		if (!bio)
-			do_bio_wait(4);
+			do_bio_wait();
 	}
 
 	bio->bi_bdev = io_info->dev;
-	bio->bi_sector = io_info->block[0];
+	bio->bi_sector = io_info->first_block;
 	bio->bi_private = io_info;
 	bio->bi_end_io = suspend_end_bio;
-	bio->bi_flags |= (1 << BIO_SUSPEND_DATA);
 	io_info->sys_struct = bio;
 
-	if (bio_add_page(bio, io_info->buffer_page, PAGE_SIZE, 0) < PAGE_SIZE) {
+	if (bio_add_page(bio, io_info->bio_page, PAGE_SIZE, 0) < PAGE_SIZE) {
 		printk("ERROR: adding page to bio at %lld\n",
-				(unsigned long long) io_info->block[0]);
+				(unsigned long long) io_info->first_block);
 		bio_put(bio);
 		return -EFAULT;
 	}
 
-	if (rw == WRITE)
+	if (io_info->writing)
 		bio_set_pages_dirty(bio);
 
 	spin_lock_irqsave(&ioinfo_busy_lock, flags);
 	list_add_tail(&io_info->list, &ioinfo_busy);
 	spin_unlock_irqrestore(&ioinfo_busy_lock, flags);
 	
-	submit_bio(rw,bio);
+	submit_bio(io_info->writing, bio);
 
-	return error;
+	return 0;
 }
 
 /* 
@@ -526,27 +355,19 @@ static int submit_batched(void)
 	unsigned long flags;
 	int num_submitted = 0;
 
+	if (running_already)
+		return 0;
+
 	running_already = 1;
 	spin_lock_irqsave(&ioinfo_submit_lock, flags);
 	while(!list_empty(&ioinfo_submit_batch)) {
 		first = list_entry(ioinfo_submit_batch.next, struct io_info,
-									list);
-
-		BUG_ON(!test_and_clear_bit(IO_AWAITING_SUBMIT, &first->flags));
-
+				list);
 		list_del_init(&first->list);
-
 		atomic_dec(&submit_batch);
-
 		spin_unlock_irqrestore(&ioinfo_submit_lock, flags);
-
-		if (test_bit(IO_AWAITING_READ, &first->flags))
-			submit(READ, first);
-		else
-			submit(WRITE, first);
-
+		submit(first);
 		spin_lock_irqsave(&ioinfo_submit_lock, flags);
-		
 		num_submitted++;
 		if (num_submitted == submit_batch_size)
 			break;
@@ -560,17 +381,15 @@ static int submit_batched(void)
 static void add_to_batch(struct io_info *io_info)
 {
 	unsigned long flags;
-
-	set_bit(IO_AWAITING_SUBMIT, &io_info->flags);
+	int waiting;
 
 	/* Put our prepared I/O struct on the batch list. */
 	spin_lock_irqsave(&ioinfo_submit_lock, flags);
 	list_add_tail(&io_info->list, &ioinfo_submit_batch);
+	waiting = atomic_add_return(1, &submit_batch);
 	spin_unlock_irqrestore(&ioinfo_submit_lock, flags);
 
-	atomic_inc(&submit_batch);
-
-	if (atomic_read(&submit_batch) >= submit_batch_size)
+	if (waiting >= submit_batch_size)
 		submit_batched();
 }
 
@@ -582,59 +401,21 @@ static void add_to_batch(struct io_info *io_info)
  */
 static struct io_info *get_io_info_struct(void)
 {
-	unsigned long newpage = 0, flags;
 	struct io_info *this = NULL;
-	int remaining = 0;
 
 	do {
-		while (atomic_read(&outstanding_io) >= MAX_OUTSTANDING_IO)
-			do_bio_wait(0);
+		while (atomic_read(&outstanding_io) >= max_outstanding_io)
+			do_bio_wait();
 
-		/* Can start a new I/O. Is there a free one? */
-		if (!list_empty(&ioinfo_free)) {
-			/* Yes. Grab it. */
-			spin_lock_irqsave(&ioinfo_free_lock, flags);
-			break;
-		}
+		this = kmalloc(sizeof(struct io_info), GFP_ATOMIC);
+	} while (!this);
 
-		/* No. Need to allocate a new page for I/O info structs. */
-		newpage = get_zeroed_page(GFP_ATOMIC);
-		if (!newpage) {
-			do_bio_wait(1);
-			continue;
-		}
-
-		suspend_message(SUSPEND_MEMORY, SUSPEND_VERBOSE, 0,
-				"[NewIOPage %lx]", newpage);
-		infopages++;
-		if (infopages > maxinfopages)
-			maxinfopages++;
-
-		/* Prepare the new page for use. */
-		this = (struct io_info *) newpage;
-		remaining = PAGE_SIZE;
-		spin_lock_irqsave(&ioinfo_free_lock, flags);
-		while (remaining >= (sizeof(struct io_info))) {
-			list_add_tail(&this->list, &ioinfo_free);
-			this = (struct io_info *) (((char *) this) + 
-					sizeof(struct io_info));
-			remaining -= sizeof(struct io_info);
-		}
-		break;
-	} while (1);
-
-	/*
-	 * We have an I/O info struct. Remove it from the free list.
-	 * It will be added to the submit or busy list later.
-	 */
-	this = list_entry(ioinfo_free.next, struct io_info, list);
-	list_del_init(&this->list);
-	spin_unlock_irqrestore(&ioinfo_free_lock, flags);
+	INIT_LIST_HEAD(&this->list);
 	return this;
 }
 
 /*
- * start_one
+ * suspend_do_io
  *
  * Description:	Prepare and start a read or write operation.
  * 		Note that we use our own buffer for reading or writing.
@@ -646,35 +427,38 @@ static struct io_info *get_io_info_struct(void)
  * 		Failure? What's that?
  * Returns:	The io_info struct created.
  */
-static struct io_info *start_one(int rw, struct submit_params *submit_info)
+static int suspend_do_io(int writing, struct block_device *bdev, long block0,
+	struct page *page, int readahead_index, int syncio)
 {
-	struct io_info *io_info = get_io_info_struct();
+	struct io_info *io_info;
 	unsigned long buffer_virt = 0;
 	char *to, *from;
-	struct page *buffer_page;
 
-	if (!io_info)
-		return NULL;
+	io_info = get_io_info_struct();
+
+	/* Done before submitting to avoid races. */
+	if (syncio)
+		waiting_on = io_info;
 
 	/* Get our local buffer */
 	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 1,
 			"Start_IO: [%p]", io_info);
 	
 	/* Copy settings to the io_info struct */
-	io_info->data_page = submit_info->page;
-	io_info->readahead_index = submit_info->readahead_index;
+	io_info->writing = writing;
+	io_info->dev = bdev;
+	io_info->first_block = block0;
+	io_info->dest_page = page;
+	io_info->readahead_index = readahead_index;
 
 	if (io_info->readahead_index == -1) {
-		while (!(buffer_virt = get_zeroed_page(GFP_ATOMIC)))
-			do_bio_wait(5);
+		while (!(buffer_virt = get_zeroed_page(GFP_ATOMIC | __GFP_NOWARN)))
+			do_bio_wait();
 
-		atomic_inc(&buffer_allocs);
 		suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
 				"[ALLOC BUFFER]->%d",
-				real_nr_free_pages());
-		buffer_page = virt_to_page(buffer_virt);
-	
-		io_info->buffer_page = buffer_page;
+				real_nr_free_pages(all_zones_mask));
+		io_info->bio_page = virt_to_page(buffer_virt);
 	} else {
 		unsigned long flags;
 		int index = io_info->readahead_index / BITS_PER_LONG;
@@ -684,95 +468,56 @@ static struct io_info *start_one(int rw, struct submit_params *submit_info)
 		clear_bit(bit, &suspend_readahead_flags[index]);
 		spin_unlock_irqrestore(&suspend_readahead_flags_lock, flags);
 
-		io_info->buffer_page = buffer_page = submit_info->page;
+		io_info->bio_page = page;
 	}
 
 	/* If writing, copy our data. The data is probably in
 	 * lowmem, but we cannot be certain. If there is no
 	 * compression/encryption, we might be passed the
 	 * actual source page's address. */
-	if (rw == WRITE) {
+	if (writing) {
 		to = (char *) buffer_virt;
-		from = kmap_atomic(io_info->data_page, KM_USER1);
+		from = kmap_atomic(page, KM_USER1);
 		memcpy(to, from, PAGE_SIZE);
 		kunmap_atomic(from, KM_USER1);
 	}
 
 	/* Submit the page */
-	get_page(buffer_page);
+	get_page(io_info->bio_page);
 	
-	io_info->dev = submit_info->dev;
-	io_info->block[0] = submit_info->block[0];
-
-	if (rw == READ)
-		set_bit(IO_AWAITING_READ, &io_info->flags);
-
 	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 1,
-			"-> (PRE BRW) %d\n",
-			real_nr_free_pages());
+		"-> (PRE BRW) %d\n", real_nr_free_pages(all_zones_mask));
 
-	if (submit_batch_size > 1)
-		add_to_batch(io_info);
+	if (syncio)
+	 	submit(io_info);
 	else
-	 	submit(rw, io_info);
+		add_to_batch(io_info);
 	
 	atomic_inc(&outstanding_io);
-	if (atomic_read(&outstanding_io) > max_outstanding_io)
-		max_outstanding_io++;
 	
-	return io_info;
-}
+	if (syncio)
+		do { do_bio_wait(); } while (waiting_on);
 
-static int suspend_do_io(int rw, 
-		struct submit_params *submit_info, int syncio)
-{
-	struct io_info *io_info;
-
-	if(!submit_info->dev) {
-		printk("Suspend_do_io: submit_info->dev is NULL!\n");
-		return 1;
-	}
-	
-	io_info = start_one(rw, submit_info);
-
-	if (!io_info) {
-		printk("Unable to allocate an io_info struct.\n");
-		return 1;
-	} else if (syncio)
-		wait_on_one_page(io_info);
-
-	/* If we were the only one, clean everything up */
-	if (!atomic_read(&outstanding_io))
-		suspend_finish_all_io();
 	return 0;
-} 
+}
 
 /* We used to use bread here, but it doesn't correctly handle
  * blocksize != PAGE_SIZE. Now we create a submit_info to get the data we
  * want and use our normal routines (synchronously).
  */
 
-static int suspend_bdev_page_io(int rw, struct block_device *bdev, long pos,
-		struct page *page)
+static int suspend_bdev_page_io(int writing, struct block_device *bdev,
+		long pos, struct page *page)
 {
-	struct submit_params submit_info;
-
-	if (!bdev)
-		return 0;
-
-	submit_info.page = page;
-	submit_info.dev = bdev;
-	submit_info.block[0] = pos;
-	submit_info.readahead_index = -1;
-	return suspend_do_io(rw, &submit_info, 1);
+	return suspend_do_io(writing, bdev, pos, page, -1, 1);
 }
 
-static unsigned long suspend_bio_memory_needed(void)
+static int suspend_bio_memory_needed(void)
 {
 	/* We want to have at least enough memory so as to have
-	 * MAX_OUTSTANDING_IO transactions on the fly at once. If we 
-	 * can to more, fine. */
-	return (MAX_OUTSTANDING_IO * (PAGE_SIZE + sizeof(struct request) +
+	 * max_outstanding_io transactions on the fly at once. If we 
+	 * can do more, fine. */
+	return (max_outstanding_io * (PAGE_SIZE + sizeof(struct request) +
 				sizeof(struct bio) + sizeof(struct io_info)));
 }
 
@@ -781,6 +526,35 @@ static void suspend_set_devinfo(struct suspend_bdev_info *info)
 	suspend_devinfo = info;
 }
 
+static void dump_block_chains(void)
+{
+	int i;
+
+	for (i = 0; i < suspend_writer_posn.num_chains; i++) {
+		struct extent *this;
+
+		printk("Chain %d:", i);
+
+		this = (suspend_writer_posn.chains + i)->first;
+
+		if (!this)
+			printk(" (Empty)");
+
+		while (this) {
+			printk(" [%lu-%lu]%s", this->minimum, this->maximum,
+					this->next ? "," : "");
+			this = this->next;
+		}
+
+		printk("\n");
+	}
+
+	for (i = 0; i < 3; i++)
+		printk("Posn %d: Chain %d, extent %d, offset %lu.\n", i,
+				suspend_writer_posn_save[i].chain_num,
+				suspend_writer_posn_save[i].extent_num,
+				suspend_writer_posn_save[i].offset);
+}
 static int forward_extra_blocks(void)
 {
 	int i;
@@ -791,6 +565,7 @@ static int forward_extra_blocks(void)
 
 	if (suspend_extent_state_eof(&suspend_writer_posn)) {
 		printk("Extent state eof.\n");
+		dump_block_chains();
 		return -ENODATA;
 	}
 
@@ -805,9 +580,8 @@ static int forward_one_page(void)
 	 * before we can know how many more blocks to skip.*/
 	suspend_extent_state_next(&suspend_writer_posn);
 
-	if (!at_start)
-		if (forward_extra_blocks())
-			return -ENODATA;
+	if (!at_start && forward_extra_blocks())
+		return -ENODATA;
 
 	if (extra_page_forward) {
 		extra_page_forward = 0;
@@ -817,55 +591,94 @@ static int forward_one_page(void)
 	return 0;
 }
 
+/* Used in reading header, to jump to 2nd page after getting 1st page
+ * direct from image header. */
 static void set_extra_page_forward(void)
 {
 	extra_page_forward = 1;
 }
 
-static int suspend_rw_page(int rw, struct page *page,
-		int readahead_index, int sync, int debug)
+static int suspend_bio_rw_page(int writing, struct page *page,
+		int readahead_index, int sync)
 {
-	int i, current_chain;
-	struct submit_params submit_params;
+	struct suspend_bdev_info *dev_info;
 
 	if (test_action_state(SUSPEND_TEST_FILTER_SPEED))
 		return 0;
 		
-	submit_params.readahead_index = readahead_index;
-	submit_params.page = page;
-	
 	if (forward_one_page()) {
 		printk("Failed to advance a page in the extent data.\n");
 		return -ENODATA;
 	}
 
-	current_chain = suspend_writer_posn.current_chain;
-	submit_params.dev = suspend_devinfo[current_chain].bdev;
-	submit_params.block[0] = suspend_writer_posn.current_offset <<
-		suspend_devinfo[current_chain].bmap_shift;
+	if (current_stream == 0 && writing &&
+		suspend_writer_posn.current_chain == suspend_writer_posn_save[2].chain_num &&
+		suspend_writer_posn.current_offset == suspend_writer_posn_save[2].offset) {
+		dump_block_chains();
+		BUG();
+	}
 
-	if (debug)
-		printk("%s: %lx:%lx.\n", rw ? "Write" : "Read",
-				(long) submit_params.dev->bd_dev,
-				(long) submit_params.block[0]);
+	dev_info = &suspend_devinfo[suspend_writer_posn.current_chain];
 
-	i = suspend_do_io(rw, &submit_params, sync);
+	return suspend_do_io(writing, dev_info->bdev,
+		suspend_writer_posn.current_offset <<
+			dev_info->bmap_shift,
+		page, readahead_index, sync);
+}
 
-	if (i)
-		return -EIO;
+static int suspend_rw_init(int writing, int stream_number)
+{
+	suspend_header_bytes_used = 0;
+
+	suspend_extent_state_restore(&suspend_writer_posn,
+			&suspend_writer_posn_save[stream_number]);
+
+	suspend_writer_buffer_posn = writing ? 0 : PAGE_SIZE;
+
+	current_stream = stream_number;
+
+	readahead_index = readahead_submit_index = -1;
+
+	pr_index = 0;
 
 	return 0;
 }
 
-static int suspend_bio_read_chunk(struct page *buffer_page, int sync)
+static void suspend_read_header_init(void)
+{
+	readahead_index = readahead_submit_index = -1;
+}
+
+static int suspend_rw_cleanup(int writing)
+{
+	if (writing && suspend_bio_rw_page(WRITE,
+			virt_to_page(suspend_writer_buffer), -1, 0))
+		return -EIO;
+
+	if (writing && current_stream == 2)
+		suspend_extent_state_save(&suspend_writer_posn,
+				&suspend_writer_posn_save[1]);
+	
+	suspend_finish_all_io();
+	
+	if (!writing)
+		while (readahead_index != readahead_submit_index) {
+			suspend_cleanup_readahead(readahead_index);
+			readahead_index++;
+			if (readahead_index == max_outstanding_io)
+				readahead_index = 0;
+		}
+
+	current_stream = 0;
+
+	return 0;
+}
+
+static int suspend_bio_read_page_with_readahead(void)
 {
 	static int last_result;
 	unsigned long *virt;
 
-	if (sync == SUSPEND_ASYNC)
-		return suspend_rw_page(READ, buffer_page, -1, sync, 0);
-
-	/* Start new readahead while we wait for our page */
 	if (readahead_index == -1) {
 		last_result = 0;
 		readahead_index = readahead_submit_index = 0;
@@ -875,8 +688,11 @@ static int suspend_bio_read_chunk(struct page *buffer_page, int sync)
 	if (last_result) {
 		/* We failed to submit a read, and have cleaned up
 		 * all the readahead previously submitted */
-		if (readahead_submit_index == readahead_index)
-			return -EPERM;
+		if (readahead_submit_index == readahead_index) {
+			abort_suspend(SUSPEND_FAILED_IO, "Failed to submit"
+				" a read and no readahead left.\n");
+			return -EIO;
+		}
 		goto wait;
 	}
 	
@@ -884,10 +700,9 @@ static int suspend_bio_read_chunk(struct page *buffer_page, int sync)
 		if (suspend_prepare_readahead(readahead_submit_index))
 			break;
 
-		last_result = suspend_rw_page(
-			READ,
+		last_result = suspend_bio_rw_page(READ,
 			suspend_readahead_pages[readahead_submit_index], 
-			readahead_submit_index, SUSPEND_ASYNC, 0);
+			readahead_submit_index, SUSPEND_ASYNC);
 		if (last_result) {
 			printk("Begin read chunk for page %d returned %d.\n",
 				readahead_submit_index, last_result);
@@ -897,7 +712,7 @@ static int suspend_bio_read_chunk(struct page *buffer_page, int sync)
 
 		readahead_submit_index++;
 
-		if (readahead_submit_index == MAX_OUTSTANDING_IO)
+		if (readahead_submit_index == max_outstanding_io)
 			readahead_submit_index = 0;
 
 	} while((!last_result) && (readahead_submit_index != readahead_index) &&
@@ -906,86 +721,34 @@ static int suspend_bio_read_chunk(struct page *buffer_page, int sync)
 wait:
 	suspend_wait_on_readahead(readahead_index);
 
-	virt = kmap_atomic(buffer_page, KM_USER1);
-	memcpy(virt, page_address(suspend_readahead_pages[readahead_index]),
-			PAGE_SIZE);
+	virt = kmap_atomic(suspend_readahead_pages[readahead_index], KM_USER1);
+	memcpy(suspend_writer_buffer, virt, PAGE_SIZE);
 	kunmap_atomic(virt, KM_USER1);
 
 	suspend_cleanup_readahead(readahead_index);
 
 	readahead_index++;
-	if (readahead_index == MAX_OUTSTANDING_IO)
+	if (readahead_index == max_outstanding_io)
 		readahead_index = 0;
 
 	return 0;
 }
 
-static int suspend_rw_init(int rw, int stream_number)
-{
-	suspend_extent_state_restore(&suspend_writer_posn,
-			&suspend_writer_posn_save[stream_number]);
-	current_stream = stream_number;
+/*
+ *
+ */
 
-	BUG_ON(!suspend_writer_posn.current_extent);
-
-	suspend_reset_io_stats();
-
-	readahead_index = readahead_submit_index = -1;
-
-	return 0;
-}
-
-static int suspend_rw_cleanup(int rw)
-{
-	if (rw == WRITE && current_stream == 2)
-		suspend_extent_state_save(&suspend_writer_posn,
-				&suspend_writer_posn_save[1]);
-	
-	suspend_finish_all_io();
-	
-	if (rw == READ) {
-		while (readahead_index != readahead_submit_index) {
-			suspend_cleanup_readahead(readahead_index);
-			readahead_index++;
-			if (readahead_index == MAX_OUTSTANDING_IO)
-				readahead_index = 0;
-		}
-	}
-
-	suspend_check_io_stats();
-
-	return 0;
-}
-
-static int suspend_write_chunk(struct page *buffer_page)
-{
-	return suspend_rw_page(WRITE, buffer_page, -1, 0, 0);
-}
-
-static int suspend_rw_header_chunk(int rw, struct suspend_module_ops *owner,
-		char *buffer, int buffer_size)
+static int suspend_rw_buffer(int writing, char *buffer, int buffer_size)
 {
 	int bytes_left = buffer_size;
-	
-	if (owner) {
-		owner->header_used += buffer_size;
-		if (owner->header_used > owner->header_requested) {
-			printk(KERN_EMERG "Suspend2 module %s is using more"
-				"header space (%lu) than it requested (%lu).\n",
-				owner->name,
-				owner->header_used,
-				owner->header_requested);
-			BUG();
-		}
-	}
 
-	/* Read a chunk of the header */
+	/* Read/write a chunk of the header */
 	while (bytes_left) {
 		char *source_start = buffer + buffer_size - bytes_left;
 		char *dest_start = suspend_writer_buffer + suspend_writer_buffer_posn;
 		int capacity = PAGE_SIZE - suspend_writer_buffer_posn;
-		char *to = rw ? dest_start : source_start;
-		char *from = rw ? source_start : dest_start;
+		char *to = writing ? dest_start : source_start;
+		char *from = writing ? source_start : dest_start;
 
 		if (bytes_left <= capacity) {
 			if (test_debug_state(SUSPEND_HEADER))
@@ -997,10 +760,10 @@ static int suspend_rw_header_chunk(int rw, struct suspend_module_ops *owner,
 			memcpy(to, from, bytes_left);
 			suspend_writer_buffer_posn += bytes_left;
 			suspend_header_bytes_used += bytes_left;
-			return rw ? 0 : buffer_size;
+			return 0;
 		}
 
-		/* Next to read the next page */
+		/* Complete this page and start a new one */
 		if (test_debug_state(SUSPEND_HEADER))
 			printk("Copy %d bytes (%d-%d) from %p to %p.\n",
 					capacity,
@@ -1011,57 +774,218 @@ static int suspend_rw_header_chunk(int rw, struct suspend_module_ops *owner,
 		bytes_left -= capacity;
 		suspend_header_bytes_used += capacity;
 
-		if (rw == READ && test_suspend_state(SUSPEND_TRY_RESUME_RD))
-			sys_read(suspend_read_fd,
-				suspend_writer_buffer, BLOCK_SIZE);
-		else {
-			if (suspend_rw_page(rw,
+		if (!writing) {
+			if (test_suspend_state(SUSPEND_TRY_RESUME_RD))
+				sys_read(suspend_read_fd,
+					suspend_writer_buffer, BLOCK_SIZE);
+			else
+				if (suspend_bio_read_page_with_readahead())
+					return -EIO;
+		} else if (suspend_bio_rw_page(WRITE,
 					virt_to_page(suspend_writer_buffer),
-					-1, !rw,
-					test_debug_state(SUSPEND_HEADER)))
+					-1, SUSPEND_ASYNC))
 				return -EIO;
-		}
 
 		suspend_writer_buffer_posn = 0;
 		suspend_cond_pause(0, NULL);
 	}
 
-	return rw ? 0 : buffer_size;
+	return 0;
 }
 
+/*
+ * suspend_bio_read_chunk
+ *
+ * Read a (possibly compressed and/or encrypted) page from the image,
+ * into buffer_page, returning it's index and the buffer size.
+ *
+ * If asynchronous I/O is requested, use readahead.
+ */
+
+static int suspend_bio_read_chunk(unsigned long *index, struct page *buffer_page,
+		unsigned int *buf_size, int sync)
+{
+	int result;
+	char *buffer_virt = kmap(buffer_page);
+
+	pr_index++;
+
+	while (!mutex_trylock(&suspend_bio_mutex))
+		do_bio_wait();
+	
+	if ((result = suspend_rw_buffer(READ, (char *) index,
+			sizeof(unsigned long)))) {
+		abort_suspend(SUSPEND_FAILED_IO,
+				"Read of index returned %d.\n", result);
+		goto out;
+	}
+
+	if ((result = suspend_rw_buffer(READ, (char *) buf_size, sizeof(int)))) {
+		abort_suspend(SUSPEND_FAILED_IO,
+				"Read of buffer size is %d.\n", result);
+		goto out;
+	}
+
+	result = suspend_rw_buffer(READ, buffer_virt, *buf_size);
+	if (result)
+		abort_suspend(SUSPEND_FAILED_IO,
+				"Read of data returned %d.\n", result);
+
+	PR_DEBUG("%d: Index %ld, %d bytes.\n", pr_index, *index, *buf_size);
+out:
+	mutex_unlock(&suspend_bio_mutex);
+	kunmap(buffer_page);
+	if (result)
+		abort_suspend(SUSPEND_FAILED_IO,
+			"Returning %d from suspend_bio_read_chunk.\n", result);
+	return result;
+}
+
+/*
+ * suspend_bio_write_chunk
+ *
+ * Write a (possibly compressed and/or encrypted) page to the image from
+ * the buffer, together with it's index and buffer size.
+ */
+
+static int suspend_bio_write_chunk(unsigned long index, struct page *buffer_page,
+		unsigned int buf_size)
+{
+	int result;
+	char *buffer_virt = kmap(buffer_page);
+
+	pr_index++;
+
+	while (!mutex_trylock(&suspend_bio_mutex))
+		do_bio_wait();
+	
+	if ((result = suspend_rw_buffer(WRITE, (char *) &index,
+					sizeof(unsigned long))))
+		goto out;
+
+	if ((result = suspend_rw_buffer(WRITE, (char *) &buf_size, sizeof(int))))
+		goto out;
+
+	result = suspend_rw_buffer(WRITE, buffer_virt, buf_size);
+
+	PR_DEBUG("%d: Index %ld, %d bytes.\n", pr_index, index, buf_size);
+out:
+	mutex_unlock(&suspend_bio_mutex);
+	kunmap(buffer_page);
+	return result;
+}
+
+/*
+ * suspend_rw_header_chunk
+ *
+ * Read or write a portion of the header.
+ */
+
+static int suspend_rw_header_chunk(int writing,
+		struct suspend_module_ops *owner,
+		char *buffer, int buffer_size)
+{
+	if (owner) {
+		owner->header_used += buffer_size;
+		if (owner->header_used > owner->header_requested) {
+			printk(KERN_EMERG "Suspend2 module %s is using more"
+				"header space (%u) than it requested (%u).\n",
+				owner->name,
+				owner->header_used,
+				owner->header_requested);
+			return buffer_size;
+		}
+	}
+
+	return suspend_rw_buffer(writing, buffer, buffer_size);
+}
+
+/*
+ * write_header_chunk_finish
+ *
+ * Flush any buffered writes in the section of the image.
+ */
 static int write_header_chunk_finish(void)
 {
-	return suspend_rw_page(WRITE,
-		virt_to_page(suspend_writer_buffer),
-		-1, 0, test_debug_state(SUSPEND_HEADER)) ? -EIO : 0;
+	return suspend_bio_rw_page(WRITE, virt_to_page(suspend_writer_buffer),
+		-1, 0) ? -EIO : 0;
+}
+
+static int suspend_bio_storage_needed(void)
+{
+	return 2 * sizeof(int);
+}
+
+static int suspend_bio_save_config_info(char *buf)
+{
+	int *ints = (int *) buf;
+	ints[0] = max_outstanding_io;
+	ints[1] = submit_batch_size;
+	return 2 * sizeof(int);
+}
+
+static void suspend_bio_load_config_info(char *buf, int size)
+{
+	int *ints = (int *) buf;
+	max_outstanding_io  = ints[0];
+	submit_batch_size = ints[1];
+}
+
+static int suspend_bio_initialise(int starting_cycle)
+{
+	suspend_writer_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
+
+	return suspend_writer_buffer ? 0 : -ENOMEM;
+}
+
+static void suspend_bio_cleanup(int finishing_cycle)
+{
+	if (suspend_writer_buffer) {
+		free_page((unsigned long) suspend_writer_buffer);
+		suspend_writer_buffer = NULL;
+	}
 }
 
 struct suspend_bio_ops suspend_bio_ops = {
 	.bdev_page_io = suspend_bdev_page_io,
-	.check_io_stats = suspend_check_io_stats,
-	.reset_io_stats = suspend_reset_io_stats,
 	.finish_all_io = suspend_finish_all_io,
-	.prepare_readahead = suspend_prepare_readahead,
-	.cleanup_readahead = suspend_cleanup_readahead,
-	.readahead_pages = suspend_readahead_pages,
-	.readahead_ready = suspend_readahead_ready,
 	.forward_one_page = forward_one_page,
 	.set_extra_page_forward = set_extra_page_forward,
 	.set_devinfo = suspend_set_devinfo,
 	.read_chunk = suspend_bio_read_chunk,
-	.write_chunk = suspend_write_chunk,
+	.write_chunk = suspend_bio_write_chunk,
 	.rw_init = suspend_rw_init,
 	.rw_cleanup = suspend_rw_cleanup,
+	.read_header_init = suspend_read_header_init,
 	.rw_header_chunk = suspend_rw_header_chunk,
 	.write_header_chunk_finish = write_header_chunk_finish,
 };
 
+static struct suspend_sysfs_data sysfs_params[] = {
+	{ SUSPEND2_ATTR("max_outstanding_io", SYSFS_RW),
+	  SYSFS_INT(&max_outstanding_io, 16, MAX_OUTSTANDING_IO, 0),
+	},
+
+	{ SUSPEND2_ATTR("submit_batch_size", SYSFS_RW),
+	  SYSFS_INT(&submit_batch_size, 16, SUBMIT_BATCH_SIZE, 0),
+	}
+};
+
 static struct suspend_module_ops suspend_blockwriter_ops = 
 {
 	.name					= "Block I/O",
 	.type					= MISC_MODULE,
+	.directory				= "block_io",
 	.module					= THIS_MODULE,
 	.memory_needed				= suspend_bio_memory_needed,
+	.storage_needed				= suspend_bio_storage_needed,
+	.save_config_info			= suspend_bio_save_config_info,
+	.load_config_info			= suspend_bio_load_config_info,
+	.initialise				= suspend_bio_initialise,
+	.cleanup				= suspend_bio_cleanup,
+
+	.sysfs_data		= sysfs_params,
+	.num_sysfs_entries	= sizeof(sysfs_params) / sizeof(struct suspend_sysfs_data),
 };
 
 static __init int suspend_block_io_load(void)
@@ -1069,6 +993,17 @@ static __init int suspend_block_io_load(void)
 	return suspend_register_module(&suspend_blockwriter_ops);
 }
 
+#ifdef CONFIG_SUSPEND2_FILE_EXPORTS
+EXPORT_SYMBOL_GPL(suspend_read_fd);
+#endif
+#if defined(CONFIG_SUSPEND2_FILE_EXPORTS) || defined(CONFIG_SUSPEND2_SWAP_EXPORTS)
+EXPORT_SYMBOL_GPL(suspend_writer_posn);
+EXPORT_SYMBOL_GPL(suspend_writer_posn_save);
+EXPORT_SYMBOL_GPL(suspend_writer_buffer);
+EXPORT_SYMBOL_GPL(suspend_writer_buffer_posn);
+EXPORT_SYMBOL_GPL(suspend_header_bytes_used);
+EXPORT_SYMBOL_GPL(suspend_bio_ops);
+#endif
 #ifdef MODULE
 static __exit void suspend_block_io_unload(void)
 {
