@@ -1,12 +1,12 @@
 /*
  * kernel/power/suspend_file.c
  *
- * Copyright 2005-2006 Nigel Cunningham <nigel@suspend2.net>
+ * Copyright (C) 2005-2007 Nigel Cunningham (nigel at suspend2 net)
  *
  * Distributed under GPLv2.
  * 
  * This file encapsulates functions for usage of a simple file as a
- * backing store. It is based upon the swapwriter, and shares the
+ * backing store. It is based upon the swapallocator, and shares the
  * same basic working. Here, though, we have nothing to do with
  * swapspace, and only one device to worry about.
  *
@@ -16,7 +16,7 @@
  *
  * and
  *
- * echo /path/to/my_file > /sys/power/suspend2/filewriter/target
+ * echo /path/to/my_file > /sys/power/suspend2/suspend_file/target
  *
  * then put what they find in /sys/power/suspend2/resume2
  * as their resume2= parameter in lilo.conf (and rerun lilo if using it).
@@ -37,6 +37,7 @@
 #include <linux/syscalls.h>
 #include <linux/namei.h>
 #include <linux/fs.h>
+#include <linux/root_dev.h>
 
 #include "suspend.h"
 #include "sysfs.h"
@@ -47,17 +48,17 @@
 #include "storage.h"
 #include "block_io.h"
 
-static struct suspend_module_ops filewriterops;
+static struct suspend_module_ops suspend_fileops;
 
 /* Details of our target.  */
 
-char filewriter_target[256];
+char suspend_file_target[256];
 static struct inode *target_inode;
 static struct file *target_file;
-static struct block_device *filewriter_target_bdev;
-static dev_t resume_dev_t;
+static struct block_device *suspend_file_target_bdev;
+static dev_t resume_file_dev_t;
 static int used_devt = 0;
-static int setting_filewriter_target = 0;
+static int setting_suspend_file_target = 0;
 static sector_t target_firstblock = 0, target_header_start = 0;
 static int target_storage_available = 0;
 static int target_claim = 0;
@@ -66,17 +67,19 @@ static char HaveImage[] = "HaveImage\n";
 static char NoImage[] =   "Suspend2\n";
 #define sig_size (sizeof(HaveImage) + 1)
 
-struct filewriter_header {
+struct suspend_file_header {
 	char sig[sig_size];
 	int resumed_before;
 	unsigned long first_header_block;
 };
 
-extern dev_t ROOT_DEV;
 extern char *__initdata root_device_name;
 
-/* Header_pages must be big enough for signature */
-static int header_pages, main_pages;
+/* Header Page Information */
+static int header_pages_allocated;
+
+/* Main Storage Pages */
+static int main_pages_allocated, main_pages_requested;
 
 #define target_is_normal_file() (S_ISREG(target_inode->i_mode))
 
@@ -90,6 +93,7 @@ enum {
 	GET_IMAGE_EXISTS,
 	INVALIDATE,
 	MARK_RESUME_ATTEMPTED,
+	UNMARK_RESUME_ATTEMPTED,
 };
 
 static void set_devinfo(struct block_device *bdev, int target_blkbits)
@@ -103,10 +107,16 @@ static void set_devinfo(struct block_device *bdev, int target_blkbits)
 	}
 }
 
-static int filewriter_storage_available(void)
+static int adjust_for_extra_pages(int unadjusted)
+{
+	return (unadjusted << PAGE_SHIFT) / (PAGE_SIZE + sizeof(unsigned long)
+			+ sizeof(int));
+}
+
+static int suspend_file_storage_available(void)
 {
 	int result = 0;
-	struct block_device *bdev=filewriter_target_bdev;
+	struct block_device *bdev=suspend_file_target_bdev;
 
 	if (!target_inode)
 		return 0;
@@ -131,7 +141,7 @@ static int filewriter_storage_available(void)
 				bdev->bd_disk->capacity) >> (PAGE_SHIFT - 9);
 	}
 
-	return result;
+	return adjust_for_extra_pages(result);
 }
 
 static int has_contiguous_blocks(int page_num)
@@ -156,14 +166,14 @@ static int size_ignoring_ignored_pages(void)
 {
 	int mappable = 0, i;
 	
-	if (target_is_normal_file()) {
-		for (i = 0; i < (target_inode->i_size >> PAGE_SHIFT) ; i++)
-			if (has_contiguous_blocks(i))
-				mappable++;
+	if (!target_is_normal_file())
+		return suspend_file_storage_available();
+
+	for (i = 0; i < (target_inode->i_size >> PAGE_SHIFT) ; i++)
+		if (has_contiguous_blocks(i))
+			mappable++;
 	
-		return mappable;
-	} else
-		return filewriter_storage_available();
+	return mappable;
 }
 
 static void __populate_block_list(int min, int max)
@@ -178,74 +188,72 @@ static void __populate_block_list(int min, int max)
 static void populate_block_list(void)
 {
 	int i;
+	int extent_min = -1, extent_max = -1, got_header = 0;
 	
 	if (block_chain.first)
 		suspend_put_extent_chain(&block_chain);
 
-	if (target_is_normal_file()) {
-		int extent_min = -1, extent_max = -1, got_header = 0;
-
-		for (i = 0;
-		     i < (target_inode->i_size >> PAGE_SHIFT);
-		     i++) {
-			sector_t new_sector;
-
-			if (!has_contiguous_blocks(i))
-				continue;
-
-			new_sector = bmap(target_inode,
-				(i * devinfo.blocks_per_page));
-
-			/* 
-			 * Ignore the first block in the file.
-			 * It gets the header.
-			 */
-			if (new_sector == target_firstblock >> devinfo.bmap_shift) {
-				got_header = 1;
-				continue;
-			}
-
-			/* 
-			 * I'd love to be able to fill in holes and resize 
-			 * files, but not yet...
-			 */
-
-			if (new_sector == extent_max + 1)
-				extent_max+= devinfo.blocks_per_page;
-			else {
-				if (extent_min > -1)
-					__populate_block_list(extent_min,
-							extent_max);
-
-				extent_min = new_sector;
-				extent_max = extent_min +
-					devinfo.blocks_per_page - 1;
-			}
-		}
-		if (extent_min > -1)
-			__populate_block_list(extent_min, extent_max);
-
-		BUG_ON(!got_header);
-	} else
+	if (!target_is_normal_file()) {
 		if (target_storage_available > 0)
 			__populate_block_list(devinfo.blocks_per_page, 
-				(min(main_pages, target_storage_available) + 1) *
+				(target_storage_available + 1) *
 			 	devinfo.blocks_per_page - 1);
+		return;
+	}
+
+	for (i = 0; i < (target_inode->i_size >> PAGE_SHIFT); i++) {
+		sector_t new_sector;
+
+		if (!has_contiguous_blocks(i))
+			continue;
+
+		new_sector = bmap(target_inode,
+		(i * devinfo.blocks_per_page));
+
+		/* 
+		 * Ignore the first block in the file.
+		 * It gets the header.
+		 */
+		if (new_sector == target_firstblock >> devinfo.bmap_shift) {
+			got_header = 1;
+			continue;
+		}
+
+		/* 
+		 * I'd love to be able to fill in holes and resize 
+		 * files, but not yet...
+		 */
+
+		if (new_sector == extent_max + 1)
+			extent_max+= devinfo.blocks_per_page;
+		else {
+			if (extent_min > -1)
+				__populate_block_list(extent_min,
+						extent_max);
+
+			extent_min = new_sector;
+			extent_max = extent_min +
+				devinfo.blocks_per_page - 1;
+		}
+	}
+
+	if (extent_min > -1)
+		__populate_block_list(extent_min, extent_max);
 }
 
-static void filewriter_cleanup(int finishing_cycle)
+static void suspend_file_cleanup(int finishing_cycle)
 {
-	if (filewriter_target_bdev) {
+	if (suspend_file_target_bdev) {
 		if (target_claim) {
-			bd_release(filewriter_target_bdev);
+			bd_release(suspend_file_target_bdev);
 			target_claim = 0;
 		}
 
 		if (used_devt) {
-			blkdev_put(filewriter_target_bdev);
+			blkdev_put(suspend_file_target_bdev);
 			used_devt = 0;
 		}
-		filewriter_target_bdev = NULL;
+		suspend_file_target_bdev = NULL;
 		target_inode = NULL;
 		set_devinfo(NULL, 0);
 		target_storage_available = 0;
@@ -266,21 +274,21 @@ static void filewriter_cleanup(int finishing_cycle)
  */
 static void reopen_resume_devt(void)
 {
-	filewriter_target_bdev = open_by_devnum(resume_dev_t, FMODE_READ);
-	if (IS_ERR(filewriter_target_bdev)) {
+	suspend_file_target_bdev = open_by_devnum(resume_file_dev_t, FMODE_READ);
+	if (IS_ERR(suspend_file_target_bdev)) {
 		printk("Got a dev_num (%lx) but failed to open it.\n",
-				(unsigned long) resume_dev_t);
+				(unsigned long) resume_file_dev_t);
 		return;
 	}
-	target_inode = filewriter_target_bdev->bd_inode;
-	set_devinfo(filewriter_target_bdev, target_inode->i_blkbits);
+	target_inode = suspend_file_target_bdev->bd_inode;
+	set_devinfo(suspend_file_target_bdev, target_inode->i_blkbits);
 }
 
-static void filewriter_get_target_info(char *target, int get_size,
+static void suspend_file_get_target_info(char *target, int get_size,
 		int resume2)
 {
 	if (target_file)
-		filewriter_cleanup(0);
+		suspend_file_cleanup(0);
 
 	if (!target || !strlen(target))
 		return;
@@ -297,54 +305,50 @@ static void filewriter_get_target_info(char *target, int get_size,
 		}
 
 		target_file = NULL;
-		resume_dev_t = name_to_dev_t(target);
-		if (!resume_dev_t) {
+		resume_file_dev_t = name_to_dev_t(target);
+		if (!resume_file_dev_t) {
+			struct kstat stat;
+			int error = vfs_stat(target, &stat);
 			printk("Open file %s returned %p and name_to_devt "
-					"failed.\n",
-					target, target_file);
-			if (!resume_dev_t) {
-				struct kstat stat;
-				int error = vfs_stat(target, &stat);
-				if (error) {
-					printk("Stating the file also failed."
-						" Nothing more we can do.\n");
-					return;
-				}
-				resume_dev_t = stat.rdev;
-			}
+					"failed.\n", target, target_file);
+			if (error)
+				printk("Stating the file also failed."
+					" Nothing more we can do.\n");
+			else
+				resume_file_dev_t = stat.rdev;
 			return;
 		}
-	     	filewriter_target_bdev = open_by_devnum(resume_dev_t, FMODE_READ);
-		if (IS_ERR(filewriter_target_bdev)) {
+
+	     	suspend_file_target_bdev = open_by_devnum(resume_file_dev_t,
+				FMODE_READ);
+		if (IS_ERR(suspend_file_target_bdev)) {
 			printk("Got a dev_num (%lx) but failed to open it.\n",
-					(unsigned long) resume_dev_t);
+					(unsigned long) resume_file_dev_t);
 			return;
 		}
 		used_devt = 1;
-		target_inode = filewriter_target_bdev->bd_inode;
+		target_inode = suspend_file_target_bdev->bd_inode;
 	} else
 		target_inode = target_file->f_mapping->host;
 
-	if (S_ISLNK(target_inode->i_mode) ||
-	    S_ISDIR(target_inode->i_mode) ||
-	    S_ISSOCK(target_inode->i_mode) ||
-	    S_ISFIFO(target_inode->i_mode)) {
-		printk("The filewriter works with regular files, character "
+	if (S_ISLNK(target_inode->i_mode) || S_ISDIR(target_inode->i_mode) ||
+	    S_ISSOCK(target_inode->i_mode) || S_ISFIFO(target_inode->i_mode)) {
+		printk("File support works with regular files, character "
 				"files and block devices.\n");
 		goto cleanup;
 	}
 
 	if (!used_devt) {
 		if (S_ISBLK(target_inode->i_mode)) {
-			filewriter_target_bdev = I_BDEV(target_inode);
-			if (!bd_claim(filewriter_target_bdev, &filewriterops))
+			suspend_file_target_bdev = I_BDEV(target_inode);
+			if (!bd_claim(suspend_file_target_bdev, &suspend_fileops))
 				target_claim = 1;
 		} else
-			filewriter_target_bdev = target_inode->i_sb->s_bdev;
-		resume_dev_t = filewriter_target_bdev->bd_dev;
+			suspend_file_target_bdev = target_inode->i_sb->s_bdev;
+		resume_file_dev_t = suspend_file_target_bdev->bd_dev;
 	}
 
-	set_devinfo(filewriter_target_bdev, target_inode->i_blkbits);
+	set_devinfo(suspend_file_target_bdev, target_inode->i_blkbits);
 
 	if (get_size)
 		target_storage_available = size_ignoring_ignored_pages();
@@ -363,7 +367,7 @@ cleanup:
 	target_storage_available = 0;
 }
 
-static int parse_signature(struct filewriter_header *header)
+static int parse_signature(struct suspend_file_header *header)
 {
 	int have_image = !memcmp(HaveImage, header->sig, sizeof(HaveImage) - 1);
 	int no_image_header = !memcmp(NoImage, header->sig, sizeof(NoImage) - 1);
@@ -385,7 +389,7 @@ static int parse_signature(struct filewriter_header *header)
 
 /* prepare_signature */
 
-static int prepare_signature(struct filewriter_header *current_header,
+static int prepare_signature(struct suspend_file_header *current_header,
 		unsigned long first_header_block)
 {
 	strncpy(current_header->sig, HaveImage, sizeof(HaveImage));
@@ -394,91 +398,113 @@ static int prepare_signature(struct filewriter_header *current_header,
 	return 0;
 }
 
-static int filewriter_storage_allocated(void)
+static int suspend_file_storage_allocated(void)
 {
-	int result;
-
 	if (!target_inode)
 		return 0;
 
-	if (target_is_normal_file()) {
-		result = (int) target_storage_available;
-	} else
-		result = header_pages + main_pages;
-
-	return result;
+	if (target_is_normal_file())
+		return (int) target_storage_available;
+	else
+		return header_pages_allocated + main_pages_requested;
 }
 
-static int filewriter_release_storage(void)
+static int suspend_file_release_storage(void)
 {
-	if ((test_action_state(SUSPEND_KEEP_IMAGE)) &&
-	     test_suspend_state(SUSPEND_NOW_RESUMING))
+	if (test_action_state(SUSPEND_KEEP_IMAGE) &&
+	    test_suspend_state(SUSPEND_NOW_RESUMING))
 		return 0;
 
 	suspend_put_extent_chain(&block_chain);
 
-	header_pages = main_pages = 0;
+	header_pages_allocated = 0;
+	main_pages_allocated = 0;
+	main_pages_requested = 0;
 	return 0;
 }
 
-static int filewriter_allocate_header_space(int space_requested)
+static int __suspend_file_allocate_storage(int main_storage_requested,
+		int header_storage);
+
+static int suspend_file_allocate_header_space(int space_requested)
 {
 	int i;
 
-	if (!block_chain.first)
-		return 0;
+	if (!block_chain.first && __suspend_file_allocate_storage(
+				main_pages_requested, space_requested)) {
+		printk("Failed to allocate space for the header.\n");
+		return -ENOSPC;
+	}
 
 	suspend_extent_state_goto_start(&suspend_writer_posn);
 	suspend_bio_ops.forward_one_page(); /* To first page */
-	
-	for (i = 0; i < space_requested; i++)
-		if (suspend_bio_ops.forward_one_page())
+
+	for (i = 0; i < space_requested; i++) {
+		if (suspend_bio_ops.forward_one_page()) {
+			printk("Out of space while seeking to allocate "
+					"header pages,\n");
+			header_pages_allocated = i;
 			return -ENOSPC;
+		}
+	}
+
+	header_pages_allocated = space_requested;
 
 	/* The end of header pages will be the start of pageset 2 */
 	suspend_extent_state_save(&suspend_writer_posn,
 			&suspend_writer_posn_save[2]);
-	header_pages = space_requested;
 	return 0;
 }
 
-static int filewriter_allocate_storage(int space_requested)
+static int suspend_file_allocate_storage(int space_requested)
 {
-	int result = 0, prev_header_pages;
-	int blocks_to_get = (space_requested << devinfo.bmap_shift) -
-		block_chain.size;
+	if (__suspend_file_allocate_storage(space_requested,
+				header_pages_allocated))
+		return -ENOSPC;
+
+	main_pages_requested = space_requested;
+	return -ENOSPC;
+}
+
+static int __suspend_file_allocate_storage(int main_space_requested,
+		int header_space_requested)
+{
+	int result = 0;
+
+	int extra_pages = DIV_ROUND_UP(main_space_requested *
+			(sizeof(unsigned long) + sizeof(int)), PAGE_SIZE);
+	int pages_to_get = main_space_requested + extra_pages +
+		header_space_requested;
+	int blocks_to_get = pages_to_get - block_chain.size;
 	
 	/* Only release_storage reduces the size */
 	if (blocks_to_get < 1)
 		return 0;
 
-	main_pages = space_requested;
-
 	populate_block_list();
 
 	suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
 		"Finished with block_chain.size == %d.\n",
 		block_chain.size);
 
-	if (block_chain.size < (header_pages + main_pages)) {
-		printk("Block chain size (%d) < header pages (%d) + main pages (%d) (=%d).\n",
-				block_chain.size,
-				header_pages, main_pages,
-				header_pages + main_pages);
+	if (block_chain.size < pages_to_get) {
+		printk("Block chain size (%d) < header pages (%d) + extra pages (%d) + main pages (%d) (=%d pages).\n",
+			block_chain.size, header_pages_allocated, extra_pages,
+			main_space_requested, pages_to_get);
 		result = -ENOSPC;
 	}
 
-	prev_header_pages = header_pages;
-	header_pages = 0;
-	filewriter_allocate_header_space(prev_header_pages);
+	main_pages_requested = main_space_requested;
+	main_pages_allocated = main_space_requested + extra_pages;
+
+	suspend_file_allocate_header_space(header_pages_allocated);
 	return result;
 }
 
-static int filewriter_write_header_init(void)
+static int suspend_file_write_header_init(void)
 {
 	suspend_extent_state_goto_start(&suspend_writer_posn);
 
-	suspend_writer_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
 	suspend_writer_buffer_posn = suspend_header_bytes_used = 0;
 
 	/* Info needed to bootstrap goes at the start of the header.
@@ -490,21 +516,21 @@ static int filewriter_write_header_init(void)
 	 * next header page by the time we go to use it.
 	 */
 
-	suspend_bio_ops.rw_header_chunk(WRITE, &filewriterops,
+	suspend_bio_ops.rw_header_chunk(WRITE, &suspend_fileops,
 			(char *) &suspend_writer_posn_save, 
 			sizeof(suspend_writer_posn_save));
 
-	suspend_bio_ops.rw_header_chunk(WRITE, &filewriterops,
+	suspend_bio_ops.rw_header_chunk(WRITE, &suspend_fileops,
 			(char *) &devinfo, sizeof(devinfo));
 
-	suspend_serialise_extent_chain(&filewriterops, &block_chain);
+	suspend_serialise_extent_chain(&suspend_fileops, &block_chain);
 	
 	return 0;
 }
 
-static int filewriter_write_header_cleanup(void)
+static int suspend_file_write_header_cleanup(void)
 {
-	struct filewriter_header *header;
+	struct suspend_file_header *header;
 
 	/* Write any unsaved data */
 	if (suspend_writer_buffer_posn)
@@ -516,23 +542,20 @@ static int filewriter_write_header_cleanup(void)
 	suspend_bio_ops.forward_one_page();
 
 	/* Adjust image header */
-	suspend_bio_ops.bdev_page_io(READ, filewriter_target_bdev,
+	suspend_bio_ops.bdev_page_io(READ, suspend_file_target_bdev,
 			target_firstblock,
 			virt_to_page(suspend_writer_buffer));
 
-	header = (struct filewriter_header *) suspend_writer_buffer;
+	header = (struct suspend_file_header *) suspend_writer_buffer;
 
 	prepare_signature(header,
 			suspend_writer_posn.current_offset <<
 			devinfo.bmap_shift);
 		
-	suspend_bio_ops.bdev_page_io(WRITE, filewriter_target_bdev,
+	suspend_bio_ops.bdev_page_io(WRITE, suspend_file_target_bdev,
 			target_firstblock,
 			virt_to_page(suspend_writer_buffer));
 
-	free_page((unsigned long) suspend_writer_buffer);
-	suspend_writer_buffer = NULL;
-	
 	suspend_bio_ops.finish_all_io();
 
 	return 0;
@@ -580,8 +603,8 @@ static int file_init(void)
 {
 	suspend_writer_buffer_posn = 0;
 
-	/* Read filewriter configuration */
-	suspend_bio_ops.bdev_page_io(READ, filewriter_target_bdev,
+	/* Read suspend_file configuration */
+	suspend_bio_ops.bdev_page_io(READ, suspend_file_target_bdev,
 			target_header_start,
 			virt_to_page((unsigned long) suspend_writer_buffer));
 	
@@ -597,7 +620,7 @@ static int file_init(void)
  * 1. Attempt to read the device specified with resume2=.
  * 2. Check the contents of the header for our signature.
  * 3. Warn, ignore, reset and/or continue as appropriate.
- * 4. If continuing, read the filewriter configuration section
+ * 4. If continuing, read the suspend_file configuration section
  *    of the header and set up block device info so we can read
  *    the rest of the header & image.
  *
@@ -607,20 +630,18 @@ static int file_init(void)
  * normally.
  */
 
-static int filewriter_read_header_init(void)
+static int suspend_file_read_header_init(void)
 {
 	int result;
 	struct block_device *tmp;
 
-	suspend_writer_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
-	
 	if (test_suspend_state(SUSPEND_TRY_RESUME_RD))
 		result = rd_init();
 	else
 		result = file_init();
 	
 	if (result) {
-		printk("Filewriter read header init: Failed to initialise "
+		printk("FileAllocator read header init: Failed to initialise "
 				"reading the first page of data.\n");
 		return result;
 	}
@@ -640,30 +661,28 @@ static int filewriter_read_header_init(void)
 	devinfo.bdev = tmp;
 	suspend_writer_buffer_posn += sizeof(devinfo);
 
+	suspend_bio_ops.read_header_init();
 	suspend_extent_state_goto_start(&suspend_writer_posn);
 	suspend_bio_ops.set_extra_page_forward();
 
 	suspend_header_bytes_used = suspend_writer_buffer_posn;
 
-	suspend_load_extent_chain(&block_chain);
-	
-	return 0;
+	return suspend_load_extent_chain(&block_chain);
 }
 
-static int filewriter_read_header_cleanup(void)
+static int suspend_file_read_header_cleanup(void)
 {
-	free_page((unsigned long) suspend_writer_buffer);
-	suspend_writer_buffer = NULL;
+	suspend_bio_ops.rw_cleanup(READ);
 	return 0;
 }
 
-static int filewriter_signature_op(int op)
+static int suspend_file_signature_op(int op)
 {
 	char *cur;
 	int result = 0, changed = 0;
-	struct filewriter_header *header;
+	struct suspend_file_header *header;
 	
-	if(filewriter_target_bdev <= 0)
+	if(suspend_file_target_bdev <= 0)
 		return -1;
 
 	cur = (char *) get_zeroed_page(GFP_ATOMIC);
@@ -673,11 +692,11 @@ static int filewriter_signature_op(int op)
 		return -ENOMEM;
 	}
 
-	suspend_bio_ops.bdev_page_io(READ, filewriter_target_bdev,
+	suspend_bio_ops.bdev_page_io(READ, suspend_file_target_bdev,
 			target_firstblock,
 			virt_to_page(cur));
 
-	header = (struct filewriter_header *) cur;
+	header = (struct suspend_file_header *) cur;
 	result = parse_signature(header);
 		
 	switch (op) {
@@ -695,10 +714,16 @@ static int filewriter_signature_op(int op)
 				changed = 1;
 			}
 			break;
+		case UNMARK_RESUME_ATTEMPTED:
+			if (result == 1) {
+				header->resumed_before = 0;
+				changed = 1;
+			}
+			break;
 	}
 
 	if (changed)
-		suspend_bio_ops.bdev_page_io(WRITE, filewriter_target_bdev,
+		suspend_bio_ops.bdev_page_io(WRITE, suspend_file_target_bdev,
 				target_firstblock,
 				virt_to_page(cur));
 
@@ -713,20 +738,20 @@ out:
  * Description:
  */
 
-static int filewriter_print_debug_stats(char *buffer, int size)
+static int suspend_file_print_debug_stats(char *buffer, int size)
 {
 	int len = 0;
 	
-	if (suspend_active_writer != &filewriterops) {
-		len = snprintf_used(buffer, size, "- Filewriter inactive.\n");
+	if (suspendActiveAllocator != &suspend_fileops) {
+		len = snprintf_used(buffer, size, "- FileAllocator inactive.\n");
 		return len;
 	}
 
-	len = snprintf_used(buffer, size, "- Filewriter active.\n");
+	len = snprintf_used(buffer, size, "- FileAllocator active.\n");
 
 	len+= snprintf_used(buffer+len, size-len, "  Storage available for image: "
 			"%ld pages.\n",
-			filewriter_storage_allocated());
+			suspend_file_storage_allocated());
 
 	return len;
 }
@@ -735,36 +760,34 @@ static int filewriter_print_debug_stats(char *buffer, int size)
  * Storage needed
  *
  * Returns amount of space in the image header required
- * for the filewriter's data.
+ * for the suspend_file's data.
  *
  * We ensure the space is allocated, but actually save the
  * data from write_header_init and therefore don't also define a
  * save_config_info routine.
  */
-static unsigned long filewriter_storage_needed(void)
+static int suspend_file_storage_needed(void)
 {
-	return sig_size + strlen(filewriter_target) + 1 +
+	return sig_size + strlen(suspend_file_target) + 1 +
 		3 * sizeof(struct extent_iterate_saved_state) +
 		sizeof(devinfo) +
 		sizeof(struct extent_chain) - 2 * sizeof(void *) +
-		(2 * sizeof(unsigned long) *
-		 (block_chain.allocs - block_chain.frees));
+		(2 * sizeof(unsigned long) * block_chain.num_extents);
 }
 
 /* 
- * filewriter_invalidate_image
+ * suspend_file_invalidate_image
  * 
  */
-static int filewriter_invalidate_image(void)
+static int suspend_file_invalidate_image(void)
 {
 	int result;
 
-	if (nr_suspends > 0)
-		filewriter_release_storage();
+	suspend_file_release_storage();
 
-	result = filewriter_signature_op(INVALIDATE);
+	result = suspend_file_signature_op(INVALIDATE);
 	if (result == 1 && !nr_suspends)
-		printk(KERN_WARNING name_suspend "Image invalidated.\n");
+		printk(KERN_WARNING "Suspend2: Image invalidated.\n");
 
 	return result;
 }
@@ -774,12 +797,12 @@ static int filewriter_invalidate_image(void)
  *
  */
 
-static int filewriter_image_exists(void)
+static int suspend_file_image_exists(void)
 {
-	if (!filewriter_target_bdev)
+	if (!suspend_file_target_bdev)
 		reopen_resume_devt();
 
-	return filewriter_signature_op(GET_IMAGE_EXISTS);
+	return suspend_file_signature_op(GET_IMAGE_EXISTS);
 }
 
 /*
@@ -788,22 +811,23 @@ static int filewriter_image_exists(void)
  * Record that we tried to resume from this image.
  */
 
-static void filewriter_mark_resume_attempted(void)
+static void suspend_file_mark_resume_attempted(int mark)
 {
-	filewriter_signature_op(MARK_RESUME_ATTEMPTED);
+	suspend_file_signature_op(mark ? MARK_RESUME_ATTEMPTED:
+		UNMARK_RESUME_ATTEMPTED);
 }
 
-static void filewriter_set_resume2(void)
+static void suspend_file_set_resume2(void)
 {
 	char *buffer = (char *) get_zeroed_page(GFP_ATOMIC);
 	char *buffer2 = (char *) get_zeroed_page(GFP_ATOMIC);
 	unsigned long sector = bmap(target_inode, 0);
 	int offset = 0;
 
-	if (filewriter_target_bdev) {
-		set_devinfo(filewriter_target_bdev, target_inode->i_blkbits);
+	if (suspend_file_target_bdev) {
+		set_devinfo(suspend_file_target_bdev, target_inode->i_blkbits);
 
-		bdevname(filewriter_target_bdev, buffer2);
+		bdevname(suspend_file_target_bdev, buffer2);
 		offset += snprintf(buffer + offset, PAGE_SIZE - offset, 
 				"/dev/%s", buffer2);
 		
@@ -812,58 +836,59 @@ static void filewriter_set_resume2(void)
 				":0x%lx", sector << devinfo.bmap_shift);
 	} else
 		offset += snprintf(buffer + offset, PAGE_SIZE - offset,
-				"%s is not a valid target.", filewriter_target);
+				"%s is not a valid target.", suspend_file_target);
 			
 	sprintf(resume2_file, "file:%s", buffer);
 
 	free_page((unsigned long) buffer);
 	free_page((unsigned long) buffer2);
 
-	suspend_attempt_to_parse_resume_device();
+	suspend_attempt_to_parse_resume_device(1);
 }
 
-static int __test_filewriter_target(char *target, int resume_time)
+static int __test_suspend_file_target(char *target, int resume_time, int quiet)
 {
-	filewriter_get_target_info(target, 0, resume_time);
-	if (filewriter_signature_op(GET_IMAGE_EXISTS) > -1) {
-		printk(name_suspend "Filewriter: File signature found.\n");
+	suspend_file_get_target_info(target, 0, resume_time);
+	if (suspend_file_signature_op(GET_IMAGE_EXISTS) > -1) {
+		if (!quiet)
+			printk("Suspend2: FileAllocator: File signature found.\n");
 		if (!resume_time)
-			filewriter_set_resume2();
+			suspend_file_set_resume2();
 		
 		suspend_bio_ops.set_devinfo(&devinfo);
 		suspend_writer_posn.chains = &block_chain;
 		suspend_writer_posn.num_chains = 1;
 
-		set_suspend_state(SUSPEND_CAN_SUSPEND);
+		if (!resume_time)
+			set_suspend_state(SUSPEND_CAN_SUSPEND);
 		return 0;
 	}
 
 	clear_suspend_state(SUSPEND_CAN_SUSPEND);
 
+	if (quiet)
+		return 1;
+
 	if (*target)
-		printk(name_suspend
-			"Filewriter: Sorry. No signature found at %s.\n",
-			target);
+		printk("Suspend2: FileAllocator: Sorry. No signature found at"
+					" %s.\n", target);
 	else
 		if (!resume_time)
-			printk(name_suspend
-				"Filewriter: Sorry. Target is not set for suspending.\n");
+			printk("Suspend2: FileAllocator: Sorry. Target is not"
+						" set for suspending.\n");
 
 	return 1;
 }
 
-static void test_filewriter_target(void)
+static void test_suspend_file_target(void)
 {
-	int cant_suspend;
-
-	setting_filewriter_target = 1;
+	setting_suspend_file_target = 1;
        	
-	printk(name_suspend "Filewriter: Testing whether you can suspend:\n");
-	cant_suspend =__test_filewriter_target(filewriter_target, 0);
-
-	printk(name_suspend "Suspending %sabled.\n",  cant_suspend ? "dis" : "en");
+	printk("Suspend2: Suspending %sabled.\n",
+			__test_suspend_file_target(suspend_file_target, 0, 1) ?
+			"dis" : "en");
 	
-	setting_filewriter_target = 0;
+	setting_suspend_file_target = 0;
 }
 
 /*
@@ -879,12 +904,13 @@ static void test_filewriter_target(void)
  * BLOCKSIZE is the logical blocksize >= SECTOR_SIZE & <= PAGE_SIZE, 
  * mod SECTOR_SIZE == 0 of the device.
  * Data is validated by attempting to read a header from the
- * location given. Failure will result in filewriter refusing to
+ * location given. Failure will result in suspend_file refusing to
  * save an image, and a reboot with correct parameters will be
  * necessary.
  */
 
-static int filewriter_parse_sig_location(char *commandline, int only_writer)
+static int suspend_file_parse_sig_location(char *commandline,
+		int only_writer, int quiet)
 {
 	char *thischar, *devstart = NULL, *colon = NULL, *at_symbol = NULL;
 	int result = -EINVAL, target_blocksize = 0;
@@ -900,7 +926,7 @@ static int filewriter_parse_sig_location(char *commandline, int only_writer)
 	 * did the initialisation successfully, assume we'll be okay when it comes
 	 * to resuming.
 	 */
-	if (filewriter_target_bdev)
+	if (suspend_file_target_bdev)
 		return 0;
 	
 	devstart = thischar = commandline;
@@ -923,19 +949,19 @@ static int filewriter_parse_sig_location(char *commandline, int only_writer)
 	}
 	
 	/* 
-	 * For the filewriter, you can be able to resume, but not suspend,
-	 * because the resume2= is set correctly, but the filewriter_target
+	 * For the suspend_file, you can be able to resume, but not suspend,
+	 * because the resume2= is set correctly, but the suspend_file_target
 	 * isn't. 
 	 *
 	 * We may have come here as a result of setting resume2 or
-	 * filewriter_target. We only test the filewriter target in the
+	 * suspend_file_target. We only test the suspend_file target in the
 	 * former case (it's already done in the later), and we do it before
 	 * setting the block number ourselves. It will overwrite the values
 	 * given on the command line if we don't.
 	 */
 
-	if (!setting_filewriter_target)
-		__test_filewriter_target(filewriter_target, 1);
+	if (!setting_suspend_file_target)
+		__test_suspend_file_target(suspend_file_target, 1, 0);
 
 	if (colon)
 		target_firstblock = (int) simple_strtoul(colon + 1, NULL, 0);
@@ -945,34 +971,34 @@ static int filewriter_parse_sig_location(char *commandline, int only_writer)
 	if (at_symbol) {
 		target_blocksize = (int) simple_strtoul(at_symbol + 1, NULL, 0);
 		if (target_blocksize & (SECTOR_SIZE - 1)) {
-			printk("Filewriter: Blocksizes are multiples of %d.\n", SECTOR_SIZE);
+			printk("FileAllocator: Blocksizes are multiples of %d.\n", SECTOR_SIZE);
 			result = -EINVAL;
 			goto out;
 		}
 	}
 	
-	printk("Suspend2 Filewriter: Testing whether you can resume:\n");
+	if (!quiet)
+		printk("Suspend2 FileAllocator: Testing whether you can resume:\n");
 
-	filewriter_get_target_info(commandline, 0, 1);
+	suspend_file_get_target_info(commandline, 0, 1);
 
-	if (!filewriter_target_bdev || IS_ERR(filewriter_target_bdev)) {
-		filewriter_target_bdev = NULL;
+	if (!suspend_file_target_bdev || IS_ERR(suspend_file_target_bdev)) {
+		suspend_file_target_bdev = NULL;
 		result = -1;
 		goto out;
 	}
 
 	if (target_blocksize)
-		set_devinfo(filewriter_target_bdev, ffs(target_blocksize));
+		set_devinfo(suspend_file_target_bdev, ffs(target_blocksize));
 
-	result = __test_filewriter_target(commandline, 1);
+	result = __test_suspend_file_target(commandline, 1, 0);
 
 out:
 	if (result)
 		clear_suspend_state(SUSPEND_CAN_SUSPEND);
-	else
-		set_suspend_state(SUSPEND_CAN_SUSPEND);
 
-	printk("Resuming %sabled.\n",  result ? "dis" : "en");
+	if (!quiet)
+		printk("Resuming %sabled.\n",  result ? "dis" : "en");
 
 	if (colon)
 		*colon = ':';
@@ -982,144 +1008,124 @@ out:
 	return result;
 }
 
-/* filewriter_save_config_info
+/* suspend_file_save_config_info
  *
  * Description:	Save the target's name, not for resume time, but for all_settings.
  * Arguments:	Buffer:		Pointer to a buffer of size PAGE_SIZE.
  * Returns:	Number of bytes used for saving our data.
  */
 
-static int filewriter_save_config_info(char *buffer)
+static int suspend_file_save_config_info(char *buffer)
 {
-	strcpy(buffer, filewriter_target);
-	return strlen(filewriter_target) + 1;
+	strcpy(buffer, suspend_file_target);
+	return strlen(suspend_file_target) + 1;
 }
 
-/* filewriter_load_config_info
+/* suspend_file_load_config_info
  *
  * Description:	Reload target's name.
  * Arguments:	Buffer:		Pointer to the start of the data.
  *		Size:		Number of bytes that were saved.
  */
 
-static void filewriter_load_config_info(char *buffer, int size)
+static void suspend_file_load_config_info(char *buffer, int size)
 {
-	strcpy(filewriter_target, buffer);
+	strcpy(suspend_file_target, buffer);
 }
 
-static int filewriter_initialise(int starting_cycle)
+static int suspend_file_initialise(int starting_cycle)
 {
-	int result = 0;
-
 	if (starting_cycle) {
-	       if (suspend_active_writer != &filewriterops)
+		if (suspendActiveAllocator != &suspend_fileops)
 			return 0;
 
-		if (!*filewriter_target) {
-			printk("Filewriter is the active writer,  but no filename has been set.\n");
+		if (starting_cycle & SYSFS_SUSPEND && !*suspend_file_target) {
+			printk("FileAllocator is the active writer,  "
+					"but no filename has been set.\n");
 			return 1;
 		}
 	}
 
-	if (filewriter_target)
-		filewriter_get_target_info(filewriter_target, starting_cycle, 0);
+	if (suspend_file_target)
+		suspend_file_get_target_info(suspend_file_target, starting_cycle, 0);
 
-	if (starting_cycle && (filewriter_image_exists() == -1)) {
+	if (starting_cycle && (suspend_file_image_exists() == -1)) {
 		printk("%s is does not have a valid signature for suspending.\n",
-				filewriter_target);
-		result = 1;
+				suspend_file_target);
+		return 1;
 	}
 
-	return result;
+	return 0;
 }
 
-static struct suspend_sysfs_data filewriter_sysfs_data[] = {
+static struct suspend_sysfs_data sysfs_params[] = {
 
 	{
 	 SUSPEND2_ATTR("target", SYSFS_RW),
-	 SYSFS_STRING(filewriter_target, 256, SYSFS_NEEDS_FOR_WRITE),
-	 .write_side_effect		= test_filewriter_target,
+	 SYSFS_STRING(suspend_file_target, 256, SYSFS_NEEDS_SM_FOR_WRITE),
+	 .write_side_effect		= test_suspend_file_target,
 	},
 
 	{
 	  SUSPEND2_ATTR("enabled", SYSFS_RW),
-	  SYSFS_INT(&filewriterops.enabled, 0, 1),
+	  SYSFS_INT(&suspend_fileops.enabled, 0, 1, 0),
 	  .write_side_effect		= attempt_to_parse_resume_device2,
 	}
 };
 
-static struct suspend_module_ops filewriterops = {
+static struct suspend_module_ops suspend_fileops = {
 	.type					= WRITER_MODULE,
-	.name					= "File Writer",
+	.name					= "File Allocator",
+	.directory				= "file",
 	.module					= THIS_MODULE,
-	.print_debug_info			= filewriter_print_debug_stats,
-	.save_config_info			= filewriter_save_config_info,
-	.load_config_info			= filewriter_load_config_info,
-	.storage_needed				= filewriter_storage_needed,
-	.initialise				= filewriter_initialise,
-	.cleanup				= filewriter_cleanup,
+	.print_debug_info			= suspend_file_print_debug_stats,
+	.save_config_info			= suspend_file_save_config_info,
+	.load_config_info			= suspend_file_load_config_info,
+	.storage_needed				= suspend_file_storage_needed,
+	.initialise				= suspend_file_initialise,
+	.cleanup				= suspend_file_cleanup,
 
-	.storage_available 	= filewriter_storage_available,
-	.storage_allocated	= filewriter_storage_allocated,
-	.release_storage	= filewriter_release_storage,
-	.allocate_header_space	= filewriter_allocate_header_space,
-	.allocate_storage	= filewriter_allocate_storage,
-	.image_exists		= filewriter_image_exists,
-	.mark_resume_attempted	= filewriter_mark_resume_attempted,
-	.write_header_init	= filewriter_write_header_init,
-	.write_header_cleanup	= filewriter_write_header_cleanup,
-	.read_header_init	= filewriter_read_header_init,
-	.read_header_cleanup	= filewriter_read_header_cleanup,
-	.invalidate_image	= filewriter_invalidate_image,
-	.parse_sig_location	= filewriter_parse_sig_location,
+	.storage_available 	= suspend_file_storage_available,
+	.storage_allocated	= suspend_file_storage_allocated,
+	.release_storage	= suspend_file_release_storage,
+	.allocate_header_space	= suspend_file_allocate_header_space,
+	.allocate_storage	= suspend_file_allocate_storage,
+	.image_exists		= suspend_file_image_exists,
+	.mark_resume_attempted	= suspend_file_mark_resume_attempted,
+	.write_header_init	= suspend_file_write_header_init,
+	.write_header_cleanup	= suspend_file_write_header_cleanup,
+	.read_header_init	= suspend_file_read_header_init,
+	.read_header_cleanup	= suspend_file_read_header_cleanup,
+	.invalidate_image	= suspend_file_invalidate_image,
+	.parse_sig_location	= suspend_file_parse_sig_location,
+
+	.sysfs_data		= sysfs_params,
+	.num_sysfs_entries	= sizeof(sysfs_params) / sizeof(struct suspend_sysfs_data),
 };
 
 /* ---- Registration ---- */
-static __init int filewriter_load(void)
+static __init int suspend_file_load(void)
 {
-	int result;
-	int i,
-	    numfiles = sizeof(filewriter_sysfs_data) / 
-		    sizeof(struct suspend_sysfs_data);
-	
-	printk("Suspend2 FileWriter loading.\n");
+	suspend_fileops.rw_init = suspend_bio_ops.rw_init;
+	suspend_fileops.rw_cleanup = suspend_bio_ops.rw_cleanup;
+	suspend_fileops.read_chunk = suspend_bio_ops.read_chunk;
+	suspend_fileops.write_chunk = suspend_bio_ops.write_chunk;
+	suspend_fileops.rw_header_chunk = suspend_bio_ops.rw_header_chunk;
 
-	filewriterops.read_chunk = suspend_bio_ops.read_chunk;
-	filewriterops.write_chunk = suspend_bio_ops.write_chunk;
-	filewriterops.rw_init = suspend_bio_ops.rw_init;
-	filewriterops.rw_cleanup = suspend_bio_ops.rw_cleanup;
-	filewriterops.rw_header_chunk =
-		suspend_bio_ops.rw_header_chunk;
-
-	if (!(result = suspend_register_module(&filewriterops))) {
-		struct kobject *kobj = make_suspend2_sysdir("filewriter");
-		for (i=0; i< numfiles; i++)
-			suspend_register_sysfs_file(kobj,
-					&filewriter_sysfs_data[i]);
-	} else
-		printk("Suspend2 FileWriter unable to register!\n");
-
-	return result;
+	return suspend_register_module(&suspend_fileops);
 }
 
 #ifdef MODULE
-static __exit void filewriter_unload(void)
+static __exit void suspend_file_unload(void)
 {
-	int i, numfiles = sizeof(filewriter_sysfs_data) /
-		sizeof(struct suspend_sysfs_data);
-
-	printk("Suspend2 FileWriter unloading.\n");
-
-	for (i=0; i< numfiles; i++)
-		suspend_unregister_sysfs_file(&filewriter_sysfs_data[i]);
-	suspend_unregister_module(&filewriterops);
+	suspend_unregister_module(&suspend_fileops);
 }
 
-module_init(filewriter_load);
-module_exit(filewriter_unload);
+module_init(suspend_file_load);
+module_exit(suspend_file_unload);
 MODULE_LICENSE("GPL");
 MODULE_AUTHOR("Nigel Cunningham");
-MODULE_DESCRIPTION("Suspend2 filewriter");
+MODULE_DESCRIPTION("Suspend2 FileAllocator");
 #else
-late_initcall(filewriter_load);
+late_initcall(suspend_file_load);
 #endif
